from .element_handle import mock_element_handle
from .frame import mock_frame
from .frame_locator import mock_frame_locator
from .js_handle import mock_js_handle
from .locator import mock_locator
from .mouse import mock_mouse
from .page import new_page

from .proxy_manager import ProxyManager
from .faker import Faker
from .playwright_mock import ChromeBrowser, FirefoxBrowser
